/*
** ** (c) 2003 - Ernst Pech
** 
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** 
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software 
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/* Codes for Output formatting
%a Abbreviated weekday name
%A Full weekday name
%b Abbreviated month name
%B Full month name
%c Date and time representation appropriate for locale
%d Day of month as decimal number (01  31)
%H Hour in 24-hour format (00  23)
%I Hour in 12-hour format (01  12)
%j Day of year as decimal number (001  366)
%m Month as decimal number (01  12)
%M Minute as decimal number (00  59)
%p Current locales A.M./P.M. indicator for 12-hour clock
%S Second as decimal number (00  59)
%U Week of year as decimal number, with Sunday as first day of week (00  53)
%w Weekday as decimal number (0  6; Sunday is 0)
%W Week of year as decimal number, with Monday as first day of week (00  53)
%x Date representation for current locale
%X Time representation for current locale
%y Year without century, as decimal number (00  99)
%YYear with century, as decimal number
%z, %Z Time-zone name or abbreviation; no characters if time zone is unknown
%% Percent sign

As in the printf function, the # flag may prefix any formatting code. In that case, the meaning of the format code is changed as follows.

Format Code Meaning 
%#a, %#A, %#b, %#B, %#p, %#X, %#z, %#Z, %#% # flag is ignored. 
%#c Long date and time representation, appropriate for current locale. For example: Tuesday, March 14, 1995, 12:41:29. 
%#x Long date representation, appropriate to current locale. For example: Tuesday, March 14, 1995. 
%#d, %#H, %#I, %#j, %#m, %#M, %#S, %#U, %#w, %#W, %#y, %#Y Remove leading zeros (if any). 
*/

#include "avisynth.h"
#include <time.h>
#include <stdio.h>


void spline(float x[], float y[], int n, float y2[]);
int splint(float xa[], float ya[], float y2a[], int n, float x, float * y, bool cubic);

AVSValue AVSChr(AVSValue args, void*,IScriptEnvironment* env )
{
    char *s = new char[2];

	s[0]=(char)(args[0].AsInt());
	s[1]=0;
    return s;
}

AVSValue AVSTime(AVSValue args, void*,IScriptEnvironment* env )
{
	time_t lt_t;
	struct tm * lt;
	time(&lt_t);
	lt = localtime (&lt_t);
    char *s = new char[50];
	strftime(s,50,args[0].AsString(""),lt);
    return s;
}

AVSValue Spline(AVSValue args, void*, IScriptEnvironment* env )
{
    float xa[256];
	float ya[256];
	float y2a[256];
	int n;
	float x,y;
	int i;
	bool cubic;

	AVSValue coordinates;

	x = args[0].AsFloat(0);
	coordinates = args[1];
	cubic = args[2].AsBool(true);

	n = coordinates.ArraySize() ;

	if (n<4 || n&1) env->ThrowError("Two few arguments for Spline");

	n=n/2;
	for (i=1; i<=n; i++) {
		xa[i] = coordinates[(i-1)*2].AsFloat(0);
		ya[i] = coordinates[(i-1)*2+1].AsFloat(0);
	}

	for (i=1; i<n; i++) {
		if (xa[i] >= xa[i+1]) env->ThrowError("Spline: all x values have to be different and in ascending order!");
	}
	
	spline(xa, ya, n, y2a);
	splint(xa, ya, y2a, n, x, &y, cubic);
	return y;
}


AVSValue String(AVSValue args, void*, IScriptEnvironment* env)
{
  if (args[0].IsString()) return args[0];
  if (args[0].IsBool()) return (args[0].AsBool()?"true":"false");
  if (args[1].Defined()) {	// a format parameter is present
		char *s = new char[80];
		if (args[0].IsFloat()) {
			sprintf(s,args[1].AsString("%f"),args[0].AsFloat());
			return s;
		}
		return "";
  } else {	// standard behaviour
	  if (args[0].IsInt()) {
		char *s = new char[12];
		return itoa(args[0].AsInt(), s, 10);
	  }
	  if (args[0].IsFloat()) {
		char *s = new char[30];
		sprintf(s,"%lf",args[0].AsFloat());
		return s;
	  }
  }
  return "";
} 

class Write : public GenericVideoFilter {

	FILE * fout;
	bool linecheck;
	bool flush;
	bool append;
	bool nofail;

	char filename[255];
	char mode[10];
	int arrsize;
	struct exp_res {
		char expression[255];
		char string[255];
	} arglist[10]; 

public:
    Write(PClip _child, const char _filename[], AVSValue args, const bool _linecheck, bool _flush, bool _append, bool _nofail, IScriptEnvironment* env):GenericVideoFilter(_child){
		linecheck = _linecheck;
		append = _append;
		flush = _flush;
		nofail = _nofail;

		strcpy(filename, _filename);
		arrsize = args.ArraySize();
		int i;
		
		for (i=0; i<10; i++) strcpy(arglist[i].string, "");

		for (i=0; i<arrsize; i++) {
			strcpy(arglist[i].expression, args[i].AsString(""));
		}

		if (append) {
			strcpy(mode, "a+t");
		} else {
			strcpy(mode, "w+t");
		}

		fout = fopen(filename, mode);	//append or purge file
		if (!fout) env->ThrowError("Write: File cannot be opened.");
		
		if (flush) fclose(fout);	//will be reopened in FileOut

		strcpy(mode, "a+t");	// in GetFrame always appending

		env->SetVar("current_frame",-1);
		if (Write::DoEval(env)) {
			Write::FileOut(env);
		}
	}

	PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);
	
	~Write() {
		if (!flush) fclose(fout);
	};
	
	bool DoEval(IScriptEnvironment* env);
	void FileOut(IScriptEnvironment* env);
};


PVideoFrame __stdcall Write::GetFrame(int n, IScriptEnvironment* env) {

	env->SetVar("current_frame",n);
	
	if (Write::DoEval(env)) {
		Write::FileOut(env);
	}

	return child->GetFrame(n, env);

};

void Write::FileOut(IScriptEnvironment* env) {
	int i;
	if (flush) {
		fout = fopen(filename, mode);
		if (!fout) env->ThrowError("Write: File cannot be opened.");
	}
	for (i= (linecheck ? 1 : 0) ; i<10; i++ ) {
		fprintf(fout, "%s", arglist[i].string );
	}
	fprintf(fout, "\n");
	if (flush) {
		fclose(fout);
	}
}

bool Write::DoEval( IScriptEnvironment* env) {
	bool keep_this_line = true;
	int i;
	AVSValue expr;
	AVSValue result;

	for (i=0; i<arrsize; i++) {
		expr = arglist[i].expression;
		
		if (linecheck && (i==0)) {
			try {
				result = env->Invoke("Eval",expr);
				if (!result.AsBool(true)) {
					keep_this_line = false;
					break;
				}
			} catch (AvisynthError error) {
//				env->ThrowError("Write: Can't eval linecheck expression!");
			}
		} else {
			try {
				result = env->Invoke("Eval",expr);
				result = env->Invoke("string",result);
				strcpy(arglist[i].string, result.AsString(""));
			} catch (AvisynthError error) {
				if (nofail) {	//display error or fallback to source expression
					strcpy(arglist[i].string, arglist[i].expression);
				} else {
					strcpy(arglist[i].string, error.msg);
				}
			}
		}
	}
	return keep_this_line;
}

AVSValue __cdecl Create_Chr(AVSValue args, void* user_data, IScriptEnvironment* env) {
	return AVSChr(args,0, env);
}
AVSValue __cdecl Create_Time(AVSValue args, void* user_data, IScriptEnvironment* env) {
	return AVSTime(args,0, env);
}

AVSValue __cdecl Create_Spline(AVSValue args, void* user_data, IScriptEnvironment* env) {
	return Spline(args,0, env);
}

AVSValue __cdecl Create_String(AVSValue args, void* user_data, IScriptEnvironment* env) {
	return String(args,0, env);
}

AVSValue __cdecl Create_Write(AVSValue args, void* user_data, IScriptEnvironment* env) {
	return new Write(args[0].AsClip(), args[1].AsString(""), args[2], 
					args[3].AsBool(false),args[4].AsBool(true),args[5].AsBool(false), args[6].AsBool(true),
					env);
}

extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env)
{
  env->AddFunction("Chr","i",Create_Chr,0);
  env->AddFunction("Time","s",Create_Time,0);
  env->AddFunction("Spline","[x]ff+[cubic]b",Create_Spline,0);
  env->AddFunction("String",".[]s",Create_String,0);
  env->AddFunction("Write","c[filename]ss+[linecheck]b[flush]b[append]b[nofail]b",Create_Write,0);
  return 0;
}


// Spline functions to generate and evaluate a natural bicubic spline
void spline(float x[], float y[], int n, float y2[])
{
	int i,k;
	float p, qn, sig, un, u[256];

	y2[1]=u[1]=0.0;

	for (i=2; i<=n-1; i++) {
		sig = (x[i] - x[i-1])/(x[i+1] - x[i-1]);
		p = sig * y2[i-1] + 2.0;
		y2[i] = (sig - 1.0) / p;
		u[i] = (y[i+1] - y[i])/(x[i+1] - x[i]) - (y[i] - y[i-1])/(x[i] - x[i-1]);
		u[i] = (6.0*u[i]/(x[i+1] - x[i-1]) - sig*u[i-1])/p;
	}
	qn=un=0.0;
	y2[n]=(un - qn*u[n-1])/(qn * y2[n-1] + 1.0);
	for (k=n-1; k>=1; k--) {
		y2[k] = y2[k] * y2[k+1] + u[k];
	}
}

int splint(float xa[], float ya[], float y2a[], int n, float x, float * y, bool cubic)
{
	int klo, khi, k;
	float h,b,a;

	klo=1;
	khi=n;
	while (khi-klo > 1) {
		k=(khi + klo) >> 1;
		if (xa[k] > x ) khi = k;
		else klo = k;
	}
	h = xa[khi] - xa[klo];
	if (h==0.0) {
		y=0;
		return -1;	// all x's have to be different
	}
	a = (xa[khi] - x)/h;
	b = (x - xa[klo])/h;

	if (cubic) {
		*y = a * ya[klo] + b*ya[khi] + ((a*a*a - a)*y2a[klo] + (b*b*b - b)*y2a[khi]) * (h*h) / 6.0;
	} else {
		*y = a * ya[klo] + b*ya[khi];
	}
	return 0;
}
